#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from bs4 import BeautifulSoup
import os
import sys
import datetime
from pyparsing import StringEnd,Literal,Word,ZeroOrMore,OneOrMore,Optional,Forward,And,Or,alphas,nums

databasestart = datetime.date(2022,1,1)
databasestop = datetime.date(2022,5,3) # one day past end
plotstart = datetime.date(2022,3,1)
plotstop = datetime.date(2022,5,3) # one day past end

# ----- translation

dictionary = {
  '（':'(',
  '）':')',
  '。':'.',
  '，':',',
  '、':',',
  '；':';',
  '：':':',
  '\u00a0':' ', # no-break space
  '\u2002':' ', # en space
  '\u3000':' ', # ideographic space
}
with open('dictionary') as f:
  for line in f:
    if '#' in line:
      line = line[:line.find('#')]
    line = line.strip()
    line = line.split(':')
    dictionary[line[0]] = line[2]

expressions = {
  # locations normally transliterated into English
  # currently not including Tibet
 'Shaan west':'Shaanxi', # province
 'Zhe river':'Zhejiang', # province
 'black dragon river':'Heilongjiang', # province
 'bless construct':'Fujian', # province
 'blue-green sea':'Qinghai', # province
 'calm emblem':'Anhui', # province
 'cloud south':'Yunnan', # province
 'distant rather':'Liaoning', # province
 'expensive state':'Guizhou', # province
 'four stream':'Sichuan', # province
 'heavy celebrate':'Chongqing', # municipality
 'lake north':'Hubei', # province
 'lake south':'Hunan', # province
 'lucky forest':'Jilin', # province
 'mountain east':'Shandong', # province
 'mountain west':'Shanxi', # province
 'new border':'Xinjiang', # autonomous region
 'north capital':'Beijing', # municipality
 'rather summer':'Ningxia', # autonomous region
 'river awake':'Jiangsu', # province
 'river west':'Jiangxi', # province
 'sea south':'Hainan', # province
 'sky sweat':'Tianjin', # municipality
 'sweet respectful':'Gansu', # province
 'upper sea':'Shanghai', # municipality
 'watercourse north':'Hebei', # province
 'watercourse south':'Henan', # province
 'wide east':'Guangdong', # province
 'wide west':'Guangxi', # autonomous region
 'force group':'Corps',
 # not as simple as transliteration
 'fragrant harbor':'Hong Kong', # special administrative region
 'inlet portal':'Macau', # special administrative region
 'inside cover old':'Inner Mongolia', # autonomous region
 'tower bay land area':'Taiwan',
 # more specific locations
 'allow thrive town':'Xuchang',
 'calm positive town':'Anyang',
 'deep furrow town':'Shenzhen',
 'Zheng state town':'Zhengzhou', # capital of Henan
 # more expressions
 'it middle':'including',
 'equal in':'all in',
 'this soil':'domestic',
 'inactive lose':'death',
 'repeated count':'cumulative count',
 'self govern area':'autonomous region',
 'direct rule town':'municipality',
 'territory outside bring into':'imported',
 'trust diagnosis sick case':'symptomatic case',
 'no disease shape affect contaminate -or':'asymptomatic case',
 'raw yield construct assume Corps':'Production and Construction Corps',
 'special distinguish go regulation area':'Special Administrative Region',
}

# ----- parsing

def skip(x):
  return And(map(Literal,x.split())).suppress()

def group(id,x):
  return x.setParseAction(lambda p:[[id]+list(p)])

number = Word(nums)

regions = [
  'Anhui',
  'Beijing',
  'Chongqing',
  'Corps',
  'Fujian',
  'Gansu',
  'Guangdong',
  'Guangxi',
  'Guizhou',
  'Hainan',
  'Hebei',
  'Heilongjiang',
  'Henan',
  'Hubei',
  'Hunan',
  'Inner Mongolia',
  'Jiangsu',
  'Jiangxi',
  'Jilin',
  'Liaoning',
  'Ningxia',
  'Qinghai',
  'Shaanxi',
  'Shandong',
  'Shanghai',
  'Shanxi',
  'Sichuan',
  'Tianjin',
  'Xinjiang',
  'Yunnan',
  'Zhejiang',
]

region = Or(Literal(r) for r in regions)
town = group('town',OneOrMore(Word(alphas+"-")))

Optional(skip('cross winter ( damage ) Ao shut ring people member'))
towncase = group('towncase',town+number+skip('case'))
regioncase = group('regioncase',region+number+skip('case'))
allintown = group('allintown',Optional(skip('all'))+skip('in')+town)
allinregion = group('allinregion',skip('(')+Optional(skip('all'))+skip('in')+region+skip(')'))

regioncasetownsplitsemi = group('regioncasetownsplit',region+number+skip('case ,')+(
  group('pertown',skip('including')+ZeroOrMore(towncase+skip(','))+towncase+skip(';'))
  |(allintown+skip(';'))
))
regioncasetownsplitparen = group('regioncasetownsplit',region+number+skip('case ,')+(
  group('pertown',skip('including')+ZeroOrMore(towncase+skip(','))+towncase+skip(')'))
  |(allintown+skip(')'))
))
allinregioncasetownsplit = group('allinregioncasetownsplit',skip('(')+Optional(skip('all'))+skip('in')+region+skip(',')+(
  group('pertown',skip('including')+ZeroOrMore(towncase+skip(','))+towncase+skip(')'))
  |(allintown+skip(')'))
))
perregionsemicolon = group('perregion',skip('(')+ZeroOrMore(regioncasetownsplitsemi)+regioncasetownsplitparen)
perregioncomma = group('perregion',skip('(')+Optional(skip('including'))+ZeroOrMore(regioncase+skip(','))+regioncase+skip(')'))
parenperregion = (
   perregionsemicolon
  |perregioncomma
  |allinregioncasetownsplit
  |allinregion
  |(skip('(')+allintown+skip(')'))
)

convertasymptomatic = group('convertasymptomatic',
  skip(', contain')+number+skip('case because asymptomatic case turn for true diagnosis sick case')+parenperregion
)

imported = group('imported',skip('imported sick case')+number+skip('case')+parenperregion+Optional(convertasymptomatic))
domestic = group('domestic',skip('domestic sick case')+number+skip('case')+parenperregion+Optional(convertasymptomatic))
allimported = group('allimported',Optional(skip('equal'))+skip('for imported')+Optional(skip('sick case'))+parenperregion+Optional(convertasymptomatic))
alldomestic = group('alldomestic',Optional(skip('equal'))+skip('for domestic')+Optional(skip('sick case'))+parenperregion+Optional(convertasymptomatic))
importedvsdomestic = (allimported|alldomestic|(skip('including')+imported+skip(';')+domestic))
importedvsdomesticcomma = (allimported|alldomestic|(skip('including')+imported+skip(',')+domestic))

deaths = group('deaths',
  skip('new increase death sick case')
  +number
  +skip('case ,')
  +skip('equal for domestic sick case ,')
  +Optional(skip('all'))
  +skip('in')
  +region
  +skip(';')
)
nodeaths = group('nodeaths',skip('no new increase death sick case .'))

period = group('period',number+skip('month')+number+skip('day')+number+skip('—')+number+skip('time'))
until = group('until',number+skip('month')+number+skip('day')+number+skip('time'))

announce = skip('31 individual province ( autonomous region , municipality ) and Xinjiang Production and Construction Corps report declare')
newsymptomatic = group('newsymptomatic',
  skip('new increase true diagnosis sick case')+number+skip('case .')
  +importedvsdomestic+skip('.')
)

importeddomesticnewformat = (
   skip('including')
  +group('imported',skip('imported')+Optional(skip('sick case'))+number+skip('case'))
  +skip(',')
  +group('domestic',skip('domestic')+Optional(skip('sick case'))+number+skip('case')+parenperregion)
)


nouncertain = group('nouncertain',skip('no new increase uncertain similar sick case .'))
uncertain = group('uncertain',skip('new increase uncertain similar sick case')+number+skip('case ,')+importedvsdomesticcomma+skip('.'))
recovery = group('recovery',
  skip('when day new increase govern recover out institution sick case')+number+skip('case ,')
  +Optional(importeddomesticnewformat+skip(','))
)
tracing = group('tracing',skip('split remove medical learn observe examine aim dense slice catch touch -or')+number+skip('people ,'))

severeup = group('severeup',skip('heavy disease sick case compare before one day increase add')+number+skip('case .'))
severedown = group('severedown',skip('heavy disease sick case compare before one day decrease little')+number+skip('case .'))
severesame = group('severesame',skip('heavy disease sick case give before one day hold flat .'))

importedstatus = group('importedstatus',
  group('sick',skip('imported now have true diagnosis sick case')+number+skip('case'))
  +(
     group('severe',skip('( including heavy disease sick case')+number+skip('case ) ,'))
     |group('nosevere',skip('( no heavy disease sick case ) ,'))
   )
  +(
     group('uncertain',skip('now have uncertain similar sick case')+number+skip('case .'))
     |group('nouncertain',skip('no now have uncertain similar sick case .'))
   )
  +group('cumulativesick',skip('cumulative count true diagnosis sick case')+number+skip('case ,'))
  +group('cumulativerecovery',skip('cumulative count govern recover out institution sick case')+number+skip('case ,'))
  +(
     skip('no death sick case .')
     |skip('cumulative count death sick case 0 case .')
   )
)

totals = group('totals',
  skip('cut to')+until+skip(',')
  +skip('occupy 31 individual province ( autonomous region , municipality ) and Xinjiang Production and Construction Corps report declare , now have')
  +group('totalsick',skip('true diagnosis sick case')+number+skip('case'))
  +Optional(group('coredecrease',skip('( core decrease')+number+skip('case )')))
  +(
     group('totalsevere',skip('( including heavy disease sick case')+number+skip('case ) ,'))
     |group('nosevere',skip('( no heavy disease sick case ) ,'))
   )
  +group('cumulativerecovery',skip('cumulative count govern recover out institution sick case')+number+skip('case ,'))
  +group('cumulativedeath',skip('cumulative count death sick case')+number+skip('case ,'))
  +group('cumulativesick',skip('cumulative count report declare true diagnosis sick case')+number+skip('case'))
  +Optional(group('coredecrease',skip('( core decrease')+number+skip('case )')))
  +skip(',')
  +(
     group('uncertain',skip('now have uncertain similar sick case')+number+skip('case .'))
     |group('nouncertain',skip('no now have uncertain similar sick case .'))
   )
  +group('cumulativetracing',skip('cumulative count pursue trace arrive dense slice catch touch -or')+number+skip('people ,'))
  +group('tracing',skip('still in medical learn observe examine aim dense slice catch touch -or')+number+skip('people .'))
)
includingimported = (
  group('includingimported',skip('(')+Optional(skip('including'))+skip('imported')+number+skip('case )'))
  |group('allimported',skip('( equal for imported )'))
  |group('noimported',skip('( no imported )'))
)

asymptomatic = group('asymptomatic',
  skip('new increase asymptomatic case')
  +number
  +skip('case')
  +(
    (group('imported',skip(', including imported')+number+skip('case ,'))
    +group('domestic',skip('domestic')+number+skip('case')+parenperregion)
    )|group('onlyimported',skip('( equal for imported )'))
  )
  +( (
       skip(';')
      +group('convertasymptomatic',skip('when day turn for true diagnosis sick case')+number+skip('case')+includingimported)
      +skip(';')
      +group('endtracing',skip('when day split remove medical learn observe examine')+number+skip('case')+includingimported)
      +skip(';')
      +group('tracing',skip('still in medical learn observe examine aim asymptomatic case')+number+skip('case')+includingimported)
     )|( # new format 2022.04.11
       skip('.')
      +group('endtracing',skip('when day split remove medical learn observe examine')+Optional(skip('aim asymptomatic case'))+number+skip('case ,')+importeddomesticnewformat)
      +skip(';')
      +group('convertasymptomatic',skip('when day turn for true diagnosis sick case')+number+skip('case')+includingimported)
      +skip(';')
      +group('tracing',skip('still in medical learn observe examine aim asymptomatic case')+number+skip('case')+includingimported)
     )
  )
  +skip('.')
)
etc = group('etc',
  skip('cumulative count collect arrive harbor inlet tower land area through report true diagnosis sick case')
  +number
  +skip('case . including ,')
  +group('hongkong',skip('Hong Kong Special Administrative Region')+number+skip('case')+Optional(skip('[1]'))+skip('( out institution')+number+skip('case , death')+number+skip('case ) ,'))
  +group('macau',skip('Macau Special Administrative Region')+number+skip('case ( out institution')+number+skip('case ) ,'))
  +group('taiwan',skip('Taiwan')+number+skip('case ( out institution')+number+skip('case , death')+number+skip('case ) .'))
)
footnotes = (
  Optional(skip('[1]')+OneOrMore(Word(alphas+nums+',.')))
  +Optional(skip('fill : for core decrease 1 case .'))
  +Optional(skip('fill : for Shandong core decrease 1 case .'))
  +Optional(skip('Jilin core increase 1 case'))
  +Optional(skip('prepare fill : core increase 2 case'))
  +Optional(skip('prepare fill : Jilin core decrease 4 case .'))
  +Optional(skip('prepare fill : Jilin core increase 1 case'))
  +Optional(skip('fill : Jilin core increase 1 case Jilin core decrease 1 case'))
  +Optional(skip('close expect , Hong Kong and Macau Special Administrative Region increase add finish asymptomatic case aim diagnosis break , elaborate feeling ask see it official net .'))
  +Optional(skip('prepare fill : replace upright after aim number writing'))
  +Optional(skip('prepare fill : each province part epidemic feeling equipment body trust rest elaborate see each province part defend raw healthy happy commission official net'))
  +Optional(skip('prepare fill : Hong Kong 4 month 21 day correct out institution number occupy enter go arrange upright , equipment body trust rest elaborate see Hong Kong defend raw department official net'))
  +skip('( fill : intermediary body pull use time , ask mark fill “ trust rest come self country home defend raw healthy happy commission member meeting official aspect net stand ” . )')
  +Optional(skip('allocation enjoy arrive')) # naturally nhc now has a link to share on baidu
)

grammar = period+skip(',')
grammar += announce
grammar += newsymptomatic
grammar += (deaths|nodeaths)
grammar += (uncertain|nouncertain)
grammar += recovery
grammar += tracing
grammar += (severeup|severedown|severesame)
grammar += importedstatus
grammar += totals
grammar += announce
grammar += asymptomatic
grammar += etc
grammar += footnotes
grammar += StringEnd()

# ----- html -> parsing -> extracting domestic cases per region

data = {}

# input: e.g., ['perregion', ['regioncase', 'Shanghai', '18'], ['regioncase', 'Guangdong', '10']]
# output: e.g., {'Shanghai':18,'Guangdong':10}
def regioncollect(data,total):
  result = {r:0 for r in regions}
  if data[0] == 'perregion':
    for R in data[1:]:
      assert R[0] in ('regioncase','regioncasetownsplit')
      assert R[1] in regions
      result[R[1]] = int(R[2])
  elif data[0] in ('allinregion','allinregioncasetownsplit'):
    assert data[1] in regions
    result[data[1]] = total
  elif data[0] == 'allintown':
    # XXX: grrr, have to figure out where towns are
    if data[1][0] == 'town' and data[1][1] in regions:
      result[data[1][1]] = total
    else:
      raise Exception('unknown town %s' % data[1])
  assert sum(result[r] for r in regions) == total
  return result

def allinregion(R,total):
  result = {r:0 for r in regions}
  assert R in regions
  result[R] = total
  return result

def parse(d,html):
  soup = BeautifulSoup(html,'html.parser')
  text = soup.find('div',{'id':'xw_box'}).get_text()
  text = text.replace('\n',' ')
  print('# original:',text)
  text = text.replace('*','')
  text = ''.join(' %s '%dictionary[c] if c in dictionary else c for c in text)
  text = text.replace('  ',' ')
  text = text.replace('  ',' ')
  text = text.replace('  ',' ')
  for e in expressions:
    text = text.replace(' %s '%e,' %s '%expressions[e])

  print('# translation:',text)

  text = text.replace('Beijing cross winter ( damage ) Ao shut ring people member 4 case','Beijing 4 case, in Paralympics')

  P = grammar.parseString(text)
  print('# parsing:',list(P))

  region_dailysymptomaticdomestic = None
  region_dailyconvertdomestic = None
  region_dailyasympdomestic = None

  for x in P:
    if x[0] == 'period':
      assert int(x[1]) == d.month
      assert int(x[2]) == d.day
      assert x[3] == '0'
      assert x[4] == '24'

    if x[0] == 'newsymptomatic':
      dailysymptomatic = int(x[1])
      for y in x[2:]:
        if y[0] == 'domestic':
          dailysymptomaticdomestic = int(y[1])
          region_dailysymptomaticdomestic = regioncollect(y[2],dailysymptomaticdomestic)
          if len(y) >= 4 and y[3][0] == 'convertasymptomatic':
            dailyconvertdomestic = int(y[3][1])
            region_dailyconvertdomestic = regioncollect(y[3][2],dailyconvertdomestic)
          else:
            # XXX: assuming that if not listed then 0
            region_dailyconvertdomestic = {r:0 for r in regions}

    if x[0] == 'asymptomatic':
      dailyasymp = int(x[1])
      for y in x[2:]:
        if y[0] == 'domestic':
          dailyasympdomestic = int(y[1])
          region_dailyasympdomestic = regioncollect(y[2],dailyasympdomestic)
        elif y[0] == 'onlyimported':
          region_dailyasympdomestic = {r:0 for r in regions}

  assert region_dailysymptomaticdomestic != None
  assert region_dailyconvertdomestic != None
  assert region_dailyasympdomestic != None

  for r in regions:
    symptomatic = region_dailysymptomaticdomestic[r]
    convert = region_dailyconvertdomestic[r]
    asymp = region_dailyasympdomestic[r]

    # the reported symptomatic numbers include positive tests _plus_ conversions
    # conversions were previously tested and asymp but are now showing symptoms
    # many news reports get this wrong (e.g., saying peak 27719 instead of 27605)
    symptomatictests = symptomatic-convert
    assert symptomatictests >= 0
    positivetests = asymp+symptomatictests

    data[d,r,'domestic','positive tests'] = positivetests
    data[d,r,'domestic','symptomatic tests'] = symptomatictests
    data[d,r,'domestic','asymptomatic tests'] = asymp
    data[d,r,'domestic','symptom development after previous asymptomic test'] = convert

    print('%d-%02d-%02d %s domestic positivetests %d symptomatictests %d asymptests %d convert %d'
      % (d.year,d.month,d.day,r,positivetests,symptomatictests,asymp,convert))

# ----- main processing loop

d = databasestart
while d != databasestop:
  fn = '截至%d月%d日24时新型冠状病毒肺炎疫情最新情况.shtml' % (d.month,d.day)
  with open('nhc/2022/'+fn) as f:
    html = f.read()
  parse(d,html)
  sys.stdout.flush()
  d += datetime.timedelta(1)

# ----- plotting

os.makedirs('graphs',exist_ok=True)

import matplotlib.pyplot as plt
import matplotlib.dates as mdates
from matplotlib.backends.backend_pdf import PdfPages
  
for r in regions:
  d = plotstart
  dend = plotstop
  
  plotx = []
  ploty = []
  while d != dend:
    plotx += [d]
    ploty += [data[d,r,'domestic','positive tests']]
    d += datetime.timedelta(1)
  
  fig,ax = plt.subplots()
  
  fig.set_size_inches(4,4)
  ax.set_xlim([plotx[0]-datetime.timedelta(1),plotx[-1]+datetime.timedelta(1)])
  
  ymax = 1.1*max(ploty)
  ax.set_ylim([0,ymax])
  
  plt.plot(plotx,ploty,color='red',marker='o')
  plt.title('%s domestic positive tests' % r)
  plt.grid(visible=True)
  
  ax.xaxis.set_major_locator(mdates.WeekdayLocator(byweekday=5))
  ax.xaxis.set_major_formatter(mdates.DateFormatter('%d %b'))
  plt.setp(ax.get_xticklabels(),rotation=90)
  fig.tight_layout()
  
  with PdfPages('graphs/domestic-positivetests-%s.pdf' % r.replace(' ','')) as pdf:
    pdf.savefig()
    plt.close()
